from pwnit.utils import ask


class Interactions:
	def __init__(self, pwntube_variable: str, tab: str):
		self.pwntube_variable: str = pwntube_variable
		self.tab: str = tab
		self.functions: list[InteractionFunction] = []
		self.menu_recvuntil: str = ""

		self.menu_recvuntil = ask("Menu recvuntil (empty to finish interactions)")
		if not self.menu_recvuntil: return

		# Functions
		while True:
			function_name = ask("Function name (empty to finish interactions)")
			if not function_name: break
			self.functions.append(InteractionFunction(function_name))

	def dump(self, tab_placeholder: str):
		result = "".join([
			f"\n{tab_placeholder}{func.dump(self.pwntube_variable, self.menu_recvuntil, tab_placeholder+self.tab)}\n"
			for func in self.functions
		])
		return result


class InteractionFunction:
	def __init__(self, name: str):
		self.name = name
		self.arguments: list[Argument] = []
		
		# Option number
		self.send_to_select = ask("Send to select it", can_skip=False)
			
		# Arguments
		while True:
			argument_name = ask("Argument name (empty to end function)")
			if not argument_name: break
			argument_sendafter = ask("Send after", can_skip=False)
			self.arguments.append(Argument(argument_name, argument_sendafter))

	def dump(self, pwntube_variable: str, menu_recvuntil: str, tab: str):
		arguments = ", ".join(arg.name for arg in self.arguments)
		menu_recvuntil = menu_recvuntil.replace("\"", "\\\"")
		send_to_select = self.send_to_select.replace("\"", "\\\"")
		send_args = [
			f'{tab}{pwntube_variable}.sendlineafter(b\"' + arg.sendafter.replace("\"", "\\\"") + f'\", {arg.name} if isinstance({arg.name}, bytes) else str({arg.name}).encode())'
			for arg in self.arguments
		]
		result = "\n".join([
			f'def {self.name}({arguments}):',
			f'{tab}{pwntube_variable}.sendlineafter(b\"{menu_recvuntil}\", b\"{send_to_select}\")',
			*send_args,
		])
		return result

class Argument:
	def __init__(self, name: str, sendafter: str) -> None:
		self.name = name
		self.sendafter = sendafter



from pathlib import Path
from pwn import ELF

from pwnit.utils import log

class Binary(ELF):
	def __init__(self, filepath: Path) -> None:
		super().__init__(filepath.expanduser(), checksec=False)

		# Create path objects
		self.path: Path = Path(self.path)
		self.debug_path: Path = self.path

		log.info(f"{type(self).__name__}: {self.path}")



from pathlib import Path
import re

class Args:
	def __init__(self) -> None:

		args = self.parse_args()

		self.remote: str | None		= args["remote"]
		self.interactions: bool		= args["interactions"]
		self.template: str | None	= args["template"]
		self.only: bool				= args["only"]
		self.libc_source: bool		= args["libc_source"]
		self.patch: Path | None		= args["patch"]
		self.seccomp: bool			= args["seccomp"]
		self.yara: Path | None		= args["yara"]

		if self.remote and not re.search(r"^[^\:]+\:\d+$", self.remote):
			raise ValueError("Remote parameter without the correct syntax '<host>:<port>'")


	def parse_args(self) -> dict[str]:
		"""Parse the arguments given to the command into a dict"""

		import argparse
		parser = argparse.ArgumentParser(
			prog="pwnit",
			description="pwnit is a tool to quickly start a pwn challenge",
		)
		parser.add_argument(
			"-r", "--remote", type=str, default=None, metavar="HOST:PORT",
			help="Specify <host>:<port>",
		)
		parser.add_argument(
			"-o", "--only", action="store_true",
			help="Do only the actions specified in args",
		)
		parser.add_argument(
			"-i", "--interactions", action="store_true",
			help="Create the interactions",
		)
		parser.add_argument(
			"-t", "--template", type=str, default=None, metavar="TAG",
			help="Create the script from the template",
		)
		parser.add_argument(
			"-p", "--patch", type=Path, default=None, metavar="PATH",
			help="Patch the executable with the specified path",
		)
		parser.add_argument(
			"--seccomp", action="store_true",
			help="Print seccomp rules if present",
		)
		parser.add_argument(
			"--yara", type=Path, default=None, metavar="RULES_FILEPATH",
			help="Check for given Yara rules file",
		)
		parser.add_argument(
			"--libc-source", action="store_true",
			help="Donwload the libc source",
		)

		return parser.parse_args().__dict__



from pwn import *

use_patched = False
binary_name = "<exe_relpath:EXE_PATH>" if (not use_patched) else "<exe_debug_relpath:EXE_DEBUG_PATH>"
exe  = ELF(binary_name, checksec=True)
libc = ELF("<libc_relpath:LIBC_PATH>", checksec=False)
context.binary = exe

ru  = lambda *x, **y: io.recvuntil(*x, **y)
rl  = lambda *x, **y: io.recvline(*x, **y)
rc  = lambda *x, **y: io.recv(*x, **y)
sla = lambda *x, **y: io.sendlineafter(*x, **y)
sa  = lambda *x, **y: io.sendafter(*x, **y)
sl  = lambda *x, **y: io.sendline(*x, **y)
sn  = lambda *x, **y: io.send(*x, **y)

if args.REMOTE:
	io = connect("<host:HOST>", "<port:PORT>")
elif args.GDB:
	io = gdb.debug(binary_name, """
		c
	""", aslr=False)
else:
	io = process(binary_name)

# <interactions>


io.interactive()



from pwnlib.term.text import blue

from pwnit.utils import log, run_command
from pwnit.exe import Exe
from pwnit.libc import Libc
from pwnit.placeholders import replace_placeholders


def run_custom_commands(
		commands: list[str],
		exe: Exe | None,
		libc: Libc | None,
		remote: str,
	) -> None:

	# For each command to execute...
	for cmd in commands:

		# Handle placeholders in command
		new_cmd = replace_placeholders(cmd, exe, libc, remote, keep_missing=False)

		# Skip command execution if some placeholder is not been substituted
		if not new_cmd:
			continue

		# Run command
		with log.progress(f"$ {blue(new_cmd)} ") as progress:
			output = run_command(new_cmd, progress, shell=True)
			if output is not None:
				progress.success()
				if output:
					log.info(output)



from pathlib import Path

from pwnit.config.config import Config
from pwnit.file_manage import recognize_exe, recognize_libs
from pwnit.exe import Exe
from pwnit.libc import Libc
from pwnit.interactions import Interactions
from pwnit.template import create_script
from pwnit.commands import run_custom_commands
from pwnit.utils import log

def main():

	# Parse args and config
	config = Config()


	# Recognize exe
	exe_path = recognize_exe(Path(".").iterdir())
	if exe_path:
		exe = Exe(exe_path)
	else:
		exe = None
		log.warning("Exe not found")

	# Recognize libc
	if (not exe) or ("libc" in exe.required_libs):
		libs_path = Path(".")
		if exe and exe.runpath:
			libs_path = Path(exe.runpath.decode())
			if not libs_path.is_dir():
				log.warning("The runpath of the exe is not a directory; fallback to the cwd")
		libcs = recognize_libs(libs_path.iterdir(), {"libc"})
		libc = Libc(libcs["libc"]) if ("libc" in libcs) else None
	else:
		libc = None

	print()


	# Do with exe
	if exe:

		# Print some info
		exe.describe()
		exe.check_functions(config.check_functions)

		# Patch
		if config.patch_path:
			exe.patch(config.patch_path, libc)

		# Analyze
		if config.seccomp:
			exe.seccomp()
		if config.yara_rules:
			exe.yara(config.yara_rules)
	
		print()


	# Do with libc
	if libc:

		# Download libc source
		if config.libc_source:
			libc.download_source()

		print()


	# Do with template
	if config.template_path:

		# Interactions
		interactions = Interactions(config.pwntube_variable, config.tab) if config.interactions else None

		# Create script
		create_script(config.template_path, config.script_path, config.remote, exe, libc, interactions)

		print()


	# Run custom commands
	run_custom_commands(config.commands, exe, libc, config.remote)



from typing import Iterable
import re
from pathlib import Path
import shutil
import requests
from pwnit.utils import ask, choose, run_command


def recognize_exe(path_list: Iterable[Path]) -> Path | None:
	"""Recognize the executable from a list of paths"""

	# Initialize potential executables list
	possible_exes: list[Path] = []

	# Loop through path list
	for file in path_list:

		# Execute file command
		filecmd_output = run_command(["file", "-bL", file], timeout=1)
		if not filecmd_output:
			continue

		# Search executable regex
		match = re.search(r"^ELF [^\,]+ executable", filecmd_output)
		if not match:
			continue

		possible_exes.append(file)

	# Return correct executable path or none
	return choose(possible_exes, "Select executable:") if possible_exes else None


def recognize_libs(path_list: Iterable[Path], libs_names: Iterable[str] = set()) -> dict[str, Path]:
	"""
	Recognize the libs from a list of paths,
	returning a dict of common names and the related path of that lib.
	It is possible to filter for common names.
	"""

	# Initialize potential libraries lists
	possible_libs: dict[str, list[Path]] = {}
	search_for = r"|".join(libs_names) if libs_names else r"[A-Za-z]+"

	# Loop through path
	for file in path_list:

		# Search libs regex
		match = re.search(rf"^({search_for})(?:[^A-Za-z].*)?\.so", file.name)
		if not match:
			continue

		# Append file to possible libs
		lib_name = match.group(1)
		possible_libs[lib_name] = possible_libs.get(lib_name, []) + [file]

	# Select actual libs and return them
	return {lib_name: choose(opts, f"Select {lib_name}:") for lib_name, opts in possible_libs.items()}


def handle_path(path: str | None) -> Path | None:
	return Path(path).expanduser() if path else None


def relative_path(path: Path) -> Path:
	return path.relative_to(Path.cwd()) if path.is_relative_to(Path.cwd()) else path


def check_file(path: Path) -> bool:
	if not path.is_file():
		if path.exists():
			raise FileExistsError(f"{path} expected to be a regular file")
		return False
	return True


def check_dir(path: Path) -> bool:
	if not path.is_dir():
		if path.exists():
			raise FileExistsError(f"{path} expected to be a directory")
		return False
	return True


def fix_if_exist(path: Path) -> Path:
	"""Check if a path exists, in case ask for a new name"""

	while path.exists():
		new_name = ask(f"{path} already exists: type another path (empty to overwrite)")
		if new_name:
			path = Path(new_name)
		else:
			if path.is_dir():
				shutil.rmtree(path)
			else:
				path.unlink()
			break
	return path


def download_file(filepath: Path, url: str) -> None:
	if not check_file(filepath):
		try:
			response = requests.get(url)
			if response.ok:
				filepath.write_text(response.text)
		except:
			pass



from pathlib import Path

from pwnit.utils import log
from pwnit.file_manage import check_file
from pwnit.exe import Exe
from pwnit.libc import Libc
from pwnit.interactions import Interactions
from pwnit.placeholders import replace_placeholders


def create_script(
		template_path: Path,
		script_path: Path,
		remote: str | None = None,
		exe: Exe | None = None,
		libc: Libc | None = None,
		interactions: Interactions | None = None,
	) -> None:

	# Handle placeholders in script path
	script_path = Path(replace_placeholders(f"{script_path}", exe, libc, remote))

	# Read template file (or script file if already exists)
	if check_file(script_path):
		content = script_path.read_text()
	elif check_file(template_path):
		content = template_path.read_text()
	else:
		log.failure("There is neither a template file nor a script file. A new script will not be created")
		return

	# Replace placeholders
	new_content = replace_placeholders(content, exe, libc, remote, interactions)

	# Write new script
	script_path.write_text(new_content)
	log.success(f"Script \'{script_path}\' updated")



from typing import TypeVar
import subprocess
from pwn import log, context, options
from pwnlib.log import Progress, Logger
from pwnlib.term.text import yellow

log_silent = context.silent


def ask(prompt: str, can_skip: bool = True) -> str:
	while True:
		received = input(f" [{yellow('?')}] {prompt} > ")
		if received or can_skip: return received
		log.warning("Can't skip")


ElementT = TypeVar("ElementT")
def choose(opts: list[ElementT], prompt: str = "Choose:", default: int | None = None) -> ElementT:
	assert opts
	if len(opts) == 1: return opts[0]
	return opts[options(f"\r [{yellow('?')}] {prompt}", list(map(str, opts)), default)]


def run_command(
		args: list | str,
		progress_log: Progress | Logger | None = log,
		**kwargs,
	) -> str | None:
	"""
	Run a command, logging out failures messages in the progress or in the log.
	This function returns the output of the command, `None` on error, and `""` on timeout.
	"""
	
	assert args
	cmd = args.split(" ")[0] if isinstance(args, str) else args[0]

	# Try executing command
	try:
		return subprocess.check_output(args, stderr=subprocess.DEVNULL, text=True, **kwargs)

	# Handle command not found
	except FileNotFoundError as err:
		if progress_log:
			progress_log.failure(f"To execute this please install {cmd}")

	# Handle interrupt
	except KeyboardInterrupt as err:
		if progress_log:
			progress_log.failure(f"{cmd} interrupted")

	# Handle errors
	except subprocess.CalledProcessError as err:
		if progress_log:
			progress_log.failure(f"{cmd} failed")
		log.debug(err)
		log.debug(err.stderr)

	# Handle timeout
	except subprocess.TimeoutExpired as err:
		log.debug(f"{cmd} timeout")
		return ""

	return None



from pathlib import Path

from pwnit.file_manage import handle_path, check_file, check_dir, download_file
from pwnit.config.args import Args
from pwnit.utils import log, choose


CONFIG_DIRPATH: Path = handle_path("~/.config/pwnit/")
CONFIG_FILEPATH = CONFIG_DIRPATH / "config.yml"


class Config:
	def __init__(self) -> None:

		# Parse args
		args = Args()

		# Read and validate config
		config: dict[str] = self.validate_config(self.read_config_file())

		# Retrieve template to use:
		templates: dict = config.get("templates", {})
		template = {}
		if args.template: # if a template is been specified in args, use that if present
			if templates:
				if args.template not in templates:
					log.failure("Speficied template isn't present in the configuration")
					args.template = choose(list(templates), "Choose template to use:")
				template = templates[args.template]
			else:
				log.failure("There are no templates in the configuration")
		elif "default" in templates: # else if a default template is present
			template = templates["default"]
		elif templates: # else if there are some templates in config
			log.warning("Default template isn't present in the configuration")
			template = templates[choose(list(templates), "Choose template to use:")]
		assert isinstance(template, dict)

		# Set config variables
		self.remote: str | None			= args.remote
		self.check_functions: list[str] = config.get("check_functions", [])
		self.patch_path: Path | None	= handle_path(args.patch or config.get("patch", None))
		self.seccomp: bool				= args.seccomp or config.get("seccomp", False)
		self.yara_rules: Path | None	= handle_path(args.yara or config.get("yara", None))
		self.libc_source: bool			= args.libc_source or config.get("libc_source", False)
		self.template_path: Path | None	= handle_path(template.get("path", None))
		self.interactions: bool			= args.interactions or template.get("interactions", False)
		self.pwntube_variable: str		= template.get("pwntube_variable", "io")
		self.tab: str					= template.get("tab", "\t")
		self.script_path: Path			= handle_path(template.get("script_path", "solve_<exe_basename:>.py"))
		self.commands: list[str]		= config.get("commands", [])

		# Handle only mode
		if args.only:
			if not args.patch: self.patch_path = None
			if not args.seccomp: self.seccomp = False
			if not args.yara: self.yara_rules = None
			if not args.libc_source: self.libc_source = False
			if not args.interactions and not args.template: self.template_path = None
			if not args.interactions: self.interactions = False
			self.commands = []


	def read_config_file(self) -> dict[str]:
		import yaml

		# Check if config file exists
		if not check_file(CONFIG_FILEPATH):

			# If config dir doesn't exists, create it
			if not check_dir(CONFIG_DIRPATH):
				CONFIG_DIRPATH.mkdir()

			# Try to download missing config files
			download_file(handle_path(CONFIG_FILEPATH), "https://raw.githubusercontent.com/Church-17/pwnit/master/resources/config.yml")
			download_file(handle_path(CONFIG_DIRPATH / "findcrypt3.rules"), "https://raw.githubusercontent.com/polymorf/findcrypt-yara/master/findcrypt3.rules")
			download_file(handle_path(CONFIG_DIRPATH / "template.py"), "https://raw.githubusercontent.com/Church-17/pwnit/master/resources/template.py")

		# Parse config file
		with open(CONFIG_FILEPATH, "r") as config_file:
			config = yaml.safe_load(config_file)

		return config
	
	def validate_config(self, config: dict[str]) -> dict[str]:
		"""Validate the schema of the config using cerberus"""

		from importlib import resources
		import json
		import jsonschema

		with resources.open_text('pwnit.config', 'config.schema.json') as f:
			schema = json.load(f)

		try:
			jsonschema.validate(config, schema)
		except jsonschema.ValidationError as err:
			log.failure(f"Invalid config: {err.message}")
			exit()

		return config



from pathlib import Path
import re
import tarfile
import requests
from pwn import libcdb
from pwnit.utils import log, log_silent
from pwnit.file_manage import handle_path, check_dir
from pwnit.binary import Binary


class Libc(Binary):
	def __init__(self, filepath: Path) -> None:
		super().__init__(filepath)

		# Initialize variables
		self.source_path: Path | None = None
		self.libc_id: str | None = None
		self.libc_version: str | None = None

		# Retrieve libc_id
		with log.progress("Libc version", "Retrieving libc ID from libc.rip...") as progress:
			with log_silent:
				libc_matches = libcdb.query_libc_rip({'buildid': self.buildid.hex()})

			if libc_matches:
				self.libc_id: str = libc_matches[0]['id']

				# Retrieve libc version from libc_id
				progress.status("Retrieving libc version from libc_id...")
				match = re.search(r"\d+(?:\.\d+)+", self.libc_id)
				if match:
					self.libc_version = match.group()

			if not self.libc_version:
				if not self.libc_id:
					log.failure("Failed to retrieve libc_id from libc.rip")
				if libc_matches == []:
					log.warning("Recognized libc is not a standard libc")

				# Retrieve libc version
				progress.status("Retrieving libc version from file...")
				match = re.search(br"release version (\d+(?:\.\d+)+)", self.path.read_bytes())
				if match:
					self.libc_version = match.group(1).decode()
				else:
					self.libc_version = None
					progress.failure("Failed to retrieve libc version")

			# Print libc version and id
			if self.libc_version:
				progress.success(f"{self.libc_version}" + (f" ({self.libc_id})" if self.libc_id else ""))

		# Download libs
		with log.progress("Retrieve associated libs", "Downloading...") as progress:
			with log_silent:
				try:
					self.libs_path = handle_path(libcdb.download_libraries(self.path))
				except requests.RequestException:
					self.libs_path = None
			if self.libs_path:
				progress.success(f"Done ({self.libs_path})")
			else:
				progress.failure("Failed to download libs")


	def download_source(self) -> None:
		"""Download the source code of this libc version"""

		with log.progress("Download libc source code") as progress:

			# Check numeric libc version
			if not self.libc_version:
				progress.failure("Missing libc version")
				return
			
			# Handle cache dir
			cache_dir = Path("~/.cache/pwnit").expanduser()
			if not check_dir(cache_dir):
				cache_dir.mkdir()

			# Check cached libc sources
			source_dirname = f"glibc-{self.libc_version}"
			for cached_source in cache_dir.iterdir():
				if source_dirname == cached_source.name:
					break

			else:
				# Get libc source archive
				url = f"http://ftpmirror.gnu.org/gnu/libc/{source_dirname}.tar.gz"
				progress.status(f"Downloading from {url}...")
				try:
					response = requests.get(url)
				except requests.RequestException:
					response = None
				if not response:
					progress.failure(f"Download from {url} failed")
					return None
				archive_path = Path(f"/tmp/{source_dirname}.tar.gz")
				archive_path.write_bytes(response.content)

				# Extract archive
				progress.status("Extracting...")
				with tarfile.open(archive_path, "r:gz") as tar:
					tar.extractall(cache_dir)

			# Return libc source
			self.source_path = cache_dir / source_dirname
			progress.success(f"{self.source_path}")



from pathlib import Path
import shutil
from pwn import libcdb
from pwnlib.term.text import red, yellow, green

from pwnit.utils import log, log_silent, run_command
from pwnit.file_manage import recognize_libs, fix_if_exist
from pwnit.binary import Binary
from pwnit.libc import Libc


class Exe(Binary):
	def __init__(self, filepath: Path) -> None:
		super().__init__(filepath)

		# Make this file executable
		self.path.chmod(0o755)

		# Initialize the runnable path
		self.runnable_path: Path | None = None
		self.set_runnable_path(self.path)

		# Retrieve required libs
		libs_paths = set()
		ldd_output = run_command(["ldd", self.path], progress_log=None, timeout=1)
		if ldd_output:
			if "statically linked" in ldd_output:
				self.statically_linked = True
			else:
				self.statically_linked = False
				libs_paths = {Path(line.strip().split(" ", 1)[0]) for line in ldd_output.split("\n") if line and ("linux-vdso" not in line)}
		elif not self.statically_linked:
			try:
				libs_paths = {Path(lib) for lib in self.libs if Path(lib) != self.path}
			except:
				log.failure("Impossible to retrieve the libs requested by the executable")

		self.required_libs: dict[str, Path] = recognize_libs({Path(lib.name) for lib in libs_paths})


	def set_runnable_path(self, path: Path) -> None:
		"""Set the exe path that correctly run without initialization errors"""

		# Return if the runnable path is been already found
		if self.runnable_path:
			return

		# Check if path is runnable without errors
		with log_silent:
			check_error = run_command([path], timeout=0.5)
		if check_error is None:
			return

		# Set runnable path
		self.runnable_path = path


	def describe(self):
		"""Print the checksec info of the executable"""

		log.info("\n".join([
			f"Arch:       {self.arch}-{self.bits}-{self.endian}",
			f"Linking:    {red('Static') if self.statically_linked else green('Dynamic')}",
			f"{self.checksec()}",
		]))


	def check_functions(self, check_functions: list[str]) -> None:
		"""Print some interesting functions used in the executable"""

		found_functions = [red(f) for f in check_functions if f in self.sym]
		if found_functions:
			log.success(f"Interesting functions: {', '.join(found_functions)}")


	def patch(self, patch_path: Path, libc: Libc | None) -> None:
		"""Patch the executable with the given libc"""

		# Return if statically linked
		if self.statically_linked:
			return

		with log.progress("Patch") as progress:

			# Handle placeholders
			progress.status("Replacing placeholders...")
			from pwnit.placeholders import replace_placeholders
			patch_path = Path(replace_placeholders(f"{patch_path}", self, libc))

			# Create debug dir
			progress.status("Creating debug directory...")
			debug_dir = patch_path.parent
			if debug_dir != Path("."):
				debug_dir = fix_if_exist(patch_path.parent)
				debug_dir.mkdir(parents=True)
			patch_path = debug_dir / patch_path.name


			progress.status("Copying and unstripping libs...")

			# Get libs names of the required libs
			missing_libs = self.required_libs.copy()
			loader_path = None

			# Copy and unstrip libc
			if libc:
				new_path = debug_dir / missing_libs["libc"]
				shutil.copy2(libc.path, new_path)
				missing_libs.pop("libc")
				with log_silent:
					try:
						libcdb.unstrip_libc(f"{new_path}")
					except:
						pass
				libc.debug_path = new_path.resolve()

			# Copy the libs from cwd
			for lib, file in recognize_libs(Path(".").iterdir(), missing_libs.keys()).items():

				# Move to debug dir, changing name as the exe expects
				new_path = debug_dir / missing_libs[lib]
				shutil.copy2(file, new_path)
				missing_libs.pop(lib)

				# Handle loader
				if lib == "ld":
					loader_path = new_path

			# Copy libs from downloaded libs
			if libc and libc.libs_path:
				libs_set = {Path(lib.name) for lib in libc.libs_path.iterdir()}
				for lib, file in missing_libs.copy().items():
					if file in libs_set:

						# Move to debug dir
						shutil.copy2(libc.libs_path / file, debug_dir)
						missing_libs.pop(lib)

						# Handle loader
						if lib == "ld":
							loader_path = debug_dir / file
				
			# Check missing libs
			if missing_libs:
				log.warning(f"Missing libs for patch: {', '.join([yellow(str(lib)) for lib in missing_libs.values()])}")

			# Run patchelf
			progress.status("Run patchelf...")
			latest_path = self.path

			# If there is a loader, patch it
			if loader_path:
				loader_path.chmod(0o755)
				cmd_output = run_command(["patchelf", "--set-interpreter", loader_path, "--output", patch_path, latest_path], progress_log=progress)
				if cmd_output is None:
					return
				latest_path = patch_path
			
			# Set runpath of executable using the latest path available
			cmd_output = run_command(["patchelf", "--set-rpath", debug_dir, "--output", patch_path, latest_path], progress_log=progress)
			if cmd_output is None:
				return

			# Change exe debug path
			self.debug_path = patch_path.resolve()
			self.set_runnable_path(self.debug_path)

			# Warning about the relative runpath and the relative ld path
			if debug_dir.is_relative_to("."):
				if loader_path:
					log.info("The patched exe can run only from the actual working directory")
				else:
					log.info("The patched exe can run with the fixed libs only from the actual working directory")


	def seccomp(self, timeout: float = 1) -> None:
		"""Print the seccomp rules if present and easily reachable"""

		# Check if exists a seccomp function
		if ("prctl" in self.sym) or any(True for function in self.sym if function.startswith("seccomp")):
			with log.progress("Seccomp", "Potential seccomp detected, analyzing...") as progress:

				# Use seccomp-tools with the runnable exe
				if self.runnable_path:
					cmd_output = run_command(["seccomp-tools", "dump", f"\'{self.runnable_path}\' </dev/null >&0 2>&0"], progress_log=progress, timeout=timeout)
					if cmd_output:
						progress.success("Found something")
						log.info(cmd_output)
					else:
						progress.success("Not found anything")
				else:
					progress.failure("The executable cannot run")


	def yara(self, yara_rules: Path) -> None:
		"""Search for pattern with yara"""

		# Check Yara rules file
		if not yara_rules.is_file():
			log.failure("Yara rules file doesn't exists. The exe will not be analyzed with yara")
			return

		# Search patterns
		import yara
		rules = yara.compile(str(yara_rules))
		matches = rules.match(str(self.path))
		if matches:
			log.success("Yara found something:")
			matches_str = []
			for match in matches:
				addresses = [instance.offset for string_match in match.strings for instance in string_match.instances]
				matches_str.append(f'{match.rule} at {", ".join(map(hex, addresses))}')
			log.info("\n".join(matches_str))



import re

from pwnit.file_manage import relative_path
from pwnit.exe import Exe
from pwnit.libc import Libc
from pwnit.interactions import Interactions


def replace_placeholders(
		text: str,
		exe: Exe | None = None,
		libc: Libc | None = None,
		remote: str | None = None,
		interactions: Interactions | None = None,
		keep_missing: bool = True,
	) -> str | None:

	# Handle host and port from remote
	host, port = remote.split(":", 1) if remote else (None, None)

	# Create substitution dictionary
	substitutions: dict[str, str | None] = {
		"exe_basename": exe.path.name if exe else None,
		"exe_relpath": f"{relative_path(exe.path)}" if exe else None,
		"exe_abspath": f"{exe.path.resolve()}" if exe else None,
		"exe_debug_basename": exe.debug_path.name if exe else None,
		"exe_debug_relpath": f"{relative_path(exe.debug_path)}" if exe else None,
		"exe_debug_abspath": f"{exe.debug_path.resolve()}" if exe else None,
		"libc_basename": libc.path.name if libc else None,
		"libc_relpath": f"{relative_path(libc.path)}" if libc else None,
		"libc_abspath": f"{libc.path.resolve()}" if libc else None,
		"libc_debug_basename": libc.debug_path.name if libc else None,
		"libc_debug_relpath": f"{relative_path(libc.debug_path)}" if libc else None,
		"libc_debug_abspath": f"{libc.debug_path.resolve()}" if libc else None,
		"libc_id": libc.libc_id if libc and libc.libc_id else None,
		"libc_version": libc.libc_version if libc and libc.libc_version else None,
		"libc_source_path": f"{libc.source_path}" if libc and libc.source_path else None,
		"remote": remote,
		"host": host,
		"port": port,
		"interactions": interactions,
	}

	# Prepare to replace interactions
	interactions_matches = re.finditer(r"([ \t]*?)(?:#.*?)?<interactions(:.*?)?>", text)

	# Define error when a placeholder is cannot be substituted
	class SubstitutionError(Exception): pass

	# Define the substitution function
	def substitute(re_match: re.Match) -> str:
		placeholder = re_match.group(1)

		# If the found regex is a correct placeholder
		if placeholder in substitutions:

			# If placeholder can be substitute with a good value, return it
			if substitutions[placeholder]:

				# Handle tabs before interactions tag
				if placeholder == "interactions":
					nonlocal interactions_matches
					return interactions.dump(next(interactions_matches).group(1))
					
				return substitutions[placeholder]

			# If there is an integrated default, use it
			elif re_match.group(2):
				return re_match.group(2)[1:]

			# If the missing can't be kept, raise an error to stop the substitutions
			elif not keep_missing:
				raise SubstitutionError
		
		# If the found regex is not a correct placeholder, or the missing can be kept, don't do anything
		return re_match.group(0)

	# Substitute placeholders handling substitution error
	try:
		new_text = re.sub(r"<(.*?)(:.*?)?>", substitute, text)
	except SubstitutionError:
		return None

	return new_text
